/***************************************************************************
 *   Copyright (C) 2005 by                                                 *
 *   Lisandro Damián Nicanor Pérez Meyer - perezmeyer en/at gmail.com      *
 *   Gustavo González - gonzalgustavo en/at gmail.com                      *
 *   Pablo Odorico  pablo.odorico en/at gmail.com                          *
 *                                                                         *
 *   Modified by Lisandro Damián Nicanor Pérez Meyer on February 2008 for  *
 *   a project for the subject "Programación de sistemas gráficos"         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "glwidget.h"

#include <QMouseEvent>
#include <QDebug>

#define POW2(a)         ((a)*(a))
#define POW3(a)         ((a)*(a)*(a))

GLWidget::GLWidget(QWidget * parent)
	: QGLWidget(parent)
{
	drawAxis = false;
	currentMethod = Constants::Bezier;
	currentFigure = Constants::Patch;
	xRot = 0;
	yRot = 0;
	zRot = 0;
}

GLWidget::~GLWidget()
{
	vertexArray->clear();
	delete vertexArray;

	qDeleteAll(holdersList.begin(), holdersList.end());
	holdersList.clear();
}

void GLWidget::setDrawAxis(bool draw)
{
	drawAxis = draw;
	update();
}

void GLWidget::setFigure(Constants::Figures figure)
{
	currentFigure = figure;
	calculateSurface();
	update();
}

void GLWidget::setMethod(Constants::Methods method)
{
	currentMethod = method;
	calculateSurface();
	update();
}

void GLWidget::setXRotation(int angle)
{
	normalizeAngle(&angle);
	if(angle != xRot)
		xRot = angle;
}

void GLWidget::setYRotation(int angle)
{
	normalizeAngle(&angle);
	if(angle != yRot)
		yRot = angle;
}

void GLWidget::setZRotation(int angle)
{
	normalizeAngle(&angle);
	if(angle != zRot)
		zRot = angle;
}

void GLWidget::initializeGL()
{
	glEnable(GL_DEPTH_TEST);

	glShadeModel(GL_SMOOTH);
	glEnable(GL_NORMALIZE);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	glHint(GL_POLYGON_SMOOTH, GL_NICEST);
	glEnable(GL_POLYGON_SMOOTH);
	glHint(GL_LINE_SMOOTH, GL_NICEST);
	//glEnable(GL_LINE_SMOOTH);

	glLineWidth(1);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	setAutoBufferSwap(false);
}

void GLWidget::paintGL()
{
	glClearColor(0.0, 0.0, 0.0, 1.0);

	// Clean
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();

	glTranslatef(0.0, 0.0, -10.0);
	glRotated(xRot / 16.0, 1.0, 0.0, 0.0);
	glRotated(yRot / 16.0, 0.0, 1.0, 0.0);
	glRotated(zRot / 16.0, 0.0, 0.0, 1.0);

	// Axis
	if(drawAxis)
	{
		glBegin(GL_LINES);
			// Red, X axis
			glColor3f(1.0, 0.0, 0.0);
			glVertex3f(0.0, 0.0, 0.0);
			glVertex3f(1.0, 0.0, 0.0);

			// Green, Y axis
			glColor3f(0.0, 1.0, 0.0);
			glVertex3f(0.0, 0.0, 0.0);
			glVertex3f(0.0, 1.0, 0.0);

			// Blue, Z axis
			glColor3f(0.0, 0.0, 1.0);
			glVertex3f(0.0, 0.0, 0.0);
			glVertex3f(0.0, 0.0, 1.0);
		glEnd();
	}

	// We enable the arrays
	glEnableClientState(GL_VERTEX_ARRAY);

	for(int currentHolder=0; currentHolder < holdersList.size(); currentHolder++)
	{
		for(int i=0; i < holdersList.at(currentHolder)->rows() - 1; i++)
		{
			vertexArray = holdersList.at(currentHolder)->getVertexArray(i);

			if(vertexArray->size() > 3)
			{
				glVertexPointer(3,GL_DOUBLE,0,vertexArray->data());

				// We draw the arrays.
				// We define a different color every three holders
				if((currentHolder % 3) == 0)
					glColor4f(1.0, 0.0, 0.0, 1.0);
				else if((currentHolder % 3) == 1)
					glColor4f(0.0, 1.0, 0.0, 1.0);
				else
					glColor4f(0.0, 0.0, 1.0, 1.0);

				glDrawArrays(GL_TRIANGLE_STRIP,0, vertexArray->size()/3);

				// To draw the wireframe, we use a white line
				glColor3f(1.0, 1.0, 1.0);
				glDrawArrays(GL_LINE_STRIP,0, vertexArray->size()/3);

			}
			delete vertexArray;
		}
	}

	swapBuffers();
}

void GLWidget::resizeGL(int width, int height)
{
	int side = qMin(width, height);
	glViewport((width - side) / 2, (height - side) / 2, side, side);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0, 1.0, 1.0, -1.0, -4.0, 150.0);
	glMatrixMode(GL_MODELVIEW);
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
	lastPos = event->pos();
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
	int dx = event->x() - lastPos.x();
	int dy = event->y() - lastPos.y();

	if (event->buttons() & Qt::LeftButton)
	{
		setXRotation(xRot + 8 * dy);
		setYRotation(yRot + 8 * dx);
	}
	else if (event->buttons() & Qt::RightButton)
	{
		setXRotation(xRot + 8 * dy);
		setZRotation(zRot + 8 * dx);
	}
	lastPos = event->pos();
	update();
}

void GLWidget::normalizeAngle(int * angle)
{
	while (*angle < 0)
		*angle += 360 * 16;
	while (*angle > 360 * 16)
		*angle -= 360 * 16;
}

Point4F GLWidget::bicubicCalculation(const ConnectivityMatrix & conMat,
                                     const int firstRow, const int firstColumn,
                                     const double u, const double v)
{
	double a(0.0), b(0.0), c(0.0), d(0.0);
	double i(0.0), j(0.0), k(0.0), l(0.0);
	if(currentMethod == Constants::Bezier)
	{
		a = -(u*u*u) + 3*u*u - 3*u +1;
		b = 3*u*u*u - 6*u*u + 3*u;
		c = -3*u*u*u + 3*u*u;
		d = u*u*u;

		i = -(v*v*v) + 3*v*v - 3*v +1;
		j = 3*v*v*v - 6*v*v + 3*v;
		k = -3*v*v*v + 3*v*v;
		l = v*v*v;
	}
	else if(currentMethod == Constants::BSplines)
	{
		a = (-(u*u*u) + 3*u*u - 3*u +1)/6.0;
		b = (3*u*u*u - 6*u*u +4)/6.0;
		c = (-3*u*u*u + 3*u*u + 3*u +1)/6.0;
		d = (u*u*u)/6.0;

		i = (-(v*v*v) + 3*v*v - 3*v +1)/6.0;
		j = (3*v*v*v - 6*v*v +4)/6.0;
		k = (-3*v*v*v + 3*v*v + 3*v +1)/6.0;
		l = (v*v*v)/6.0;
	}

	Point4F e = a * conMat.getPoint(firstRow, firstColumn) +
	            b * conMat.getPoint(firstRow + 1, firstColumn) +
	            c * conMat.getPoint(firstRow + 2, firstColumn) +
	            d * conMat.getPoint(firstRow + 3, firstColumn);
	Point4F f = a * conMat.getPoint(firstRow, firstColumn + 1) +
	            b * conMat.getPoint(firstRow + 1, firstColumn + 1) +
	            c * conMat.getPoint(firstRow + 2, firstColumn + 1) +
	            d * conMat.getPoint(firstRow + 3, firstColumn + 1);
	Point4F g = a * conMat.getPoint(firstRow, firstColumn + 2) +
	            b * conMat.getPoint(firstRow + 1, firstColumn + 2) +
	            c * conMat.getPoint(firstRow + 2, firstColumn + 2) +
	            d * conMat.getPoint(firstRow + 3, firstColumn + 2);
	Point4F h = a * conMat.getPoint(firstRow, firstColumn + 3) +
	            b * conMat.getPoint(firstRow + 1, firstColumn + 3) +
	            c * conMat.getPoint(firstRow + 2, firstColumn + 3) +
	            d * conMat.getPoint(firstRow + 3, firstColumn + 3);

	double x = e.x()*i + f.x()*j + g.x()*k + h.x()*l;
	double y = e.y()*i + f.y()*j + g.y()*k + h.y()*l;
	double z = e.z()*i + f.z()*j + g.z()*k + h.z()*l;

	return Point4F(x, y, z);
}

void GLWidget::calculateSurface()
{
	// Clear all the previous data
	qDeleteAll(holdersList.begin(), holdersList.end());
	holdersList.clear();

	DataHolder * holder = 0;
	ConnectivityMatrix * conMatrix = 0;

	if(currentFigure == Constants::Patch)
		conMatrix = &patchConMat;
	else if(currentFigure == Constants::Tube)
		conMatrix = &tubeConMat;
	else if(currentFigure == Constants::SemiTube)
		conMatrix = &semiTubeConMat;
	else if(currentFigure == Constants::Torus)
		conMatrix = &torusConMat;

	double uStep = 0.0;
	double vStep = 0.0;

	/*
	  We must iterate between 0 and rows-4 and columns-4, as the algorythm takes
	  4 points in each run.
	*/
	for(int row=0; row <= conMatrix->getRows() - 4; row++)
	{
		for(int column=0; column <= conMatrix->getColumns() - 4; column++)
		{
			holder = new DataHolder(numberOfSteps + 1, numberOfSteps + 1);

			for(int u=0; u <= numberOfSteps; u++)
			{
				uStep = (double)u/(double)numberOfSteps;

				for(int v=0; v <= numberOfSteps; v++)
				{
					vStep = (double)v/(double)numberOfSteps;
					holder->addData(u, v,
					                bicubicCalculation(*conMatrix, row, column,
					                                   uStep, vStep));
				}

			}
			holdersList.append(holder);
		}
	}
}




