/***************************************************************************
 *   Copyright (C) 2005 by                                                 *
 *   Lisandro Damián Nicanor Pérez Meyer - perezmeyer en/at gmail.com      *
 *   Gustavo González - gonzalgustavo en/at gmail.com                      *
 *   Pablo Odorico  pablo.odorico en/at gmail.com                          *
 *   Originally designed for QAntenna                                      *
 *   http://qantenna.sourceforge.net/                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef GLWIDGET_H
#define GLWIDGET_H

#include <QtOpenGL/QGLWidget>
#include <QList>
#include <QVector>
#include <QPoint>
#include "constants.h"
#include "connectivitymatrix.h"
#include "patch.h"
#include "tube.h"
#include "semitube.h"
#include "torus.h"
#include "point4f.h"
#include "dataholder.h"

/**
	This class inherits from QGLWidget, and it is used to diaplay the OGL view
	in the program. It also takes cares of some events, wich it then passes to
	the class camera.
*/
class GLWidget : public QGLWidget
{
	Q_OBJECT

public:
	GLWidget(QWidget * parent = 0);
	~GLWidget();

public slots:
	/// Sets the axis to be drawn or not
	void setDrawAxis(bool draw = true);
	/// Set the figure to draw
	void setFigure(Constants::Figures figure);
	/// Set the method to draw
	void setMethod(Constants::Methods method);

	/// Sets the X rotation angle
	void setXRotation(int angle);
	/// Sets the Y rotation angle
	void setYRotation(int angle);
	/// Sets the Z rotation angle
	void setZRotation(int angle);

protected:
	void initializeGL();
	void paintGL();
	void resizeGL(int width, int height);
	void mousePressEvent(QMouseEvent *event);
	void mouseMoveEvent(QMouseEvent *event);

private:
	/// Normalize the rotation angle
	void normalizeAngle(int * angle);
	/// Bicubic calculation
	Point4F bicubicCalculation(const ConnectivityMatrix & conMat,
	                           const int firstRow, const int firstColumn,
	                           const double u, const double v);

	/// Calculate the surface
	void calculateSurface();

	/// X rotation angle
	int xRot;
	/// Y rotation angle
	int yRot;
	/// Z rotation angle
	int zRot;
	/// Last position of the mouse to calculate rotations
	QPoint lastPos;

	/// State of the axis
	bool drawAxis;

	/// Figure to draw
	int currentFigure;

	/// Method to draw
	int currentMethod;

	/// Step of u and v
	static const int numberOfSteps = 10;

	// Surfaces connectivity matrices
	/// The patch surface connectivity matrix
	Patch patchConMat;
	/// The tube surface connectivity matrix
	Tube tubeConMat;
	/// The semi tube surface connectivity matrix
	SemiTube semiTubeConMat;
	/// Torus surface connectivity matrix
	Torus torusConMat;

	/// Vertex array
	QVector<double> * vertexArray;

	/// List of holders
	QList< DataHolder * > holdersList;
};

#endif // GLWIDGET_H

